package com.brtchip.io;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import jssc.SerialPortEvent;
import jssc.SerialPortEventListener;
import jssc.SerialPortException;

public class JSSCSerialPort extends SerialPort {
		
	private jssc.SerialPort serialPort;
	private BlockingQueue<Byte> readq;
	private ArrayList<SerialPortListener> txListeners;
	private ArrayList<SerialPortListener> rxListeners;
		
	public JSSCSerialPort() 
	{
		serialPort = null;
		txListeners = new ArrayList<SerialPortListener>();
		rxListeners = new ArrayList<SerialPortListener>();
	}
	
	/** Return all available serial ports
	 * 
	 * @return An array of unopened serial ports
	 */
	public static String[] listPorts() 
	{
		ArrayList<String> availablePorts = new ArrayList<String>();
		String[] allPorts = jssc.SerialPortList.getPortNames();
		
		for (String port:allPorts)
		{
			jssc.SerialPort sp = new jssc.SerialPort(port);
			try {
				sp.openPort();
				sp.closePort();
				
				availablePorts.add(port);
				
				//System.out.println("\""+port+"\" Available");
			} catch (SerialPortException ex) {
				//System.out.println("\""+port+"\" Unavailable");
			} finally {
			}
		}
		
		return availablePorts.toArray(new String[availablePorts.size()]);
	}

	@Override
	public void open(String portname, int baud, DATABITS databits,
			STOPBITS stopbits, PARITY parity) 
	{
		serialPort = new jssc.SerialPort(portname);
		
		try {
			serialPort.openPort();
			
			int _dataBits, _stopBits, _parity;
			
			switch (databits)
			{
			case DATABITS_5: _dataBits = jssc.SerialPort.DATABITS_5; break;
			case DATABITS_6: _dataBits = jssc.SerialPort.DATABITS_6; break;
			case DATABITS_7: _dataBits = jssc.SerialPort.DATABITS_7; break;
			default:
			case DATABITS_8: _dataBits = jssc.SerialPort.DATABITS_8; break;
			}
			
			switch (stopbits)
			{
			default:
			case STOPBITS_1: _stopBits = jssc.SerialPort.STOPBITS_1; break;
			case STOPBITS_1_5: _stopBits = jssc.SerialPort.STOPBITS_1_5; break;
			case STOPBITS_2: _stopBits = jssc.SerialPort.STOPBITS_2; break;
			}
			
			switch (parity)
			{
			case PARITY_EVEN: _parity = jssc.SerialPort.PARITY_EVEN; break;
			case PARITY_MARK: _parity = jssc.SerialPort.PARITY_MARK; break;
			default:
			case PARITY_NONE: _parity = jssc.SerialPort.PARITY_NONE; break;
			case PARITY_ODD: _parity = jssc.SerialPort.PARITY_ODD; break;
			case PARITY_SPACE: _parity = jssc.SerialPort.PARITY_SPACE; break;
			}
			
			serialPort.setParams(baud, _dataBits, _stopBits, _parity);
			
			readq = new LinkedBlockingQueue<Byte>();
			
			serialPort.setEventsMask(jssc.SerialPort.MASK_RXCHAR + jssc.SerialPort.MASK_CTS + jssc.SerialPort.MASK_BREAK  
					+ jssc.SerialPort.MASK_DSR + jssc.SerialPort.MASK_RING);
			serialPort.addEventListener(new SerialPortReader(serialPort, readq, rxListeners));
		} catch (SerialPortException ex) {
			ex.printStackTrace();
		}

	}

	@Override
	public void close() {
		if (serialPort != null && serialPort.isOpened())
		{
			try {
				serialPort.closePort();
			} catch (SerialPortException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} finally {
				serialPort = null;
			}
		}
		
		clear();
		readq = null;

	}

	@Override
	public byte read() {
		byte b = 0;
		try {
			if (readq != null)
				b = readq.take();
		} catch (InterruptedException ex) {
			
		}
		return b;
	}

	@Override
	public byte read(int timeout) throws TimeoutException {
		Byte b = 0;
		try {
			b = readq.poll(timeout, TimeUnit.MILLISECONDS);
			if (b == null)
				throw new TimeoutException("SerialPort.read: timeout expired");
		} catch (InterruptedException ex) {
			
		}		
		return b;
	}

	@Override
	public void write(byte b) {
		try {
			serialPort.writeByte(b);
			for (SerialPortListener l:txListeners)
			{
				l.byteTransfered(b);
			}
		} catch (SerialPortException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	@Override
	public void write(String s) {
		try {
			if (s != null)
			{
				serialPort.writeString(s);
				for (byte b:s.getBytes())
				{
					for (SerialPortListener l:txListeners)
					{
						l.byteTransfered(b);
					}
				}
			}
		} catch (SerialPortException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	@Override
	public int available() {
		int sz = 0;
		if (readq != null)
			sz = readq.size();
		return sz;
	}

	@Override
	public void clear() {
		if (readq != null)
			readq.clear();
	}

	@Override
	public void flush() {
	}

	@Override
	public void addReceiveListener(SerialPortListener spl) {
		rxListeners.add(spl);
	}

	@Override
	public void addTransmitListener(SerialPortListener spl) {
		txListeners.add(spl);
	}
	
	
	/** Serial Port Event Reader
	 * 
	 */
	static class SerialPortReader implements SerialPortEventListener 
	{
		private final jssc.SerialPort serialPort;
		private final BlockingQueue<Byte> queue;
		private final List<SerialPortListener> listeners;
		
		public SerialPortReader(jssc.SerialPort serialPort, BlockingQueue<Byte> queue, List<SerialPortListener> listeners) {
			this.serialPort = serialPort;
			this.queue = queue;
			this.listeners = listeners;
		}
		
		@Override
		public void serialEvent(SerialPortEvent ev) {
			if (ev.isRXCHAR())
			{
				try {
					int x = ev.getEventValue();
					byte buffer[] = serialPort.readBytes(x);
					for (byte b:buffer)
					{
						queue.put(b);
						for(SerialPortListener l:listeners)
							l.byteTransfered(b);
					}
				} catch (InterruptedException ex) {
					ex.printStackTrace();
				} catch (SerialPortException ex) {
					ex.printStackTrace();
				}
			}
			else if (ev.isBREAK())
			{
				for(SerialPortListener l:listeners)
					l.onBreak();
			}
			else if (ev.isCTS())
			{
				for(SerialPortListener l:listeners)
					l.onCTS();				
			}
			else if (ev.isDSR())
			{
				for(SerialPortListener l:listeners)
					l.onDSR();				
			}
			else if (ev.isRING())
			{
				for(SerialPortListener l:listeners)
					l.onRing();
			}
			
		}
	}

}
